#!/usr/bin/ruby

require "config"
require "util"
require "cache"
require "node"
require "nodelist"

class ServerCGI
  def initialize
    $conf["debugid"] = "serv" # debug
    @head = ""
    @body = ""
  end

  attr_accessor :head, :body

  #
  # main routine
  #
  def execute(path, input)
    File.umask($conf["umask"])

    @head = "Content-type: text/plain\r\n" +

    @body = ""

    if ENV["REQUEST_METHOD"] !~ /\A(GET|HEAD)\Z/ then
      Util.debug("method ============ ", ENV["REQUEST_METHOD"]) ## debug
      return
    end

    if path == "" then
      begin
        File.open($conf["motd"], "r") {|f|
          while f.read(1024, buf) do
            body << buf
          end
        }
      rescue
        nil
      end
    end

    # ping
    if path == "ping" then
      @body << "PONG\n" << ENV["REMOTE_ADDR"] << "\n"

    # node
    elsif path == "node" then
      node = NodeList.random()
      if node then
        node = node.toString()
        @body << node << "\n"
      end

    # join
    elsif path =~ /^join\/([^:]*):(\d+)(.*)/ then
      host, nport, path = $1, $2, $3
      host = ENV["REMOTE_ADDR"].dup if host == ""

      node = Node.new(host, nport, path)
      if checkAddress(host, ENV["REMOTE_ADDR"]) && node.ping() then

        if NodeList.include(node) ||
          (NodeList.count() < $conf["nodes"]) then
          NodeList.add(node)
          @body << "WELCOME\n"
        else
          suggest = NodeList.random()
          NodeList.remove(suggest)
          suggest.bye()
          @body << "WELCOME\n" << suggest.toString() << "\n"
          NodeList.add(node)
        end
      end

    # bye
    elsif path =~ /^bye\/([^:]*):(\d+)(.*)/ then
      host, port, path = $1, $2, $3
      host = ENV["REMOTE_ADDR"].dup if host == ""
      node = Node.new(host, port, path)
      if checkAddress(host, ENV["REMOTE_ADDR"]) then
        NodeList.remove(node)
        @body << "BYEBYE\n"
      end

    # have
    elsif path =~ /^have\/([0-9A-Za-z_]+)$/ then
      file = $1
      if FileTest.size?("#{$conf["datadir"]}/#{file}.dat") then
        @body << "YES\n"
      else
        @body << "NO\n"
      end

    # get / head
    elsif path =~ /^(get|head)\/([0-9A-Za-z_]+)\/([-0-9A-Za-z\/]*)$/ then
      command, file, stamp = $1, $2, $3
      lines = (IO.readlines("#{$conf["datadir"]}/#{file}.dat") rescue [])

      if lines.size > 0 then
        items = lines[-1].split(/<>/)
        laststamp = items[0]
        if stamp == "" then
          s, e = 0, laststamp
        elsif stamp =~ /^(\d+)$/ then
          s, e = $1, $1
        elsif stamp =~ /^-(\d+)$/ then
          s, e = 0, $1
        elsif stamp =~ /^(\d+)-$/ then
          s, e = $1, laststamp
        elsif stamp =~ /^(\d+)-(\d+)$/ then
          s, e = $1, $2
        elsif stamp =~ /^(\d+)\/([0-9A-Za-z]+)/ then
          s, e, id = $1, $1, $2
        end

        if s then
          lines.each {|line|
            items = line.split(/<>/)
            next if (items.size <= 2)
  
            line = "#{items[0]}<>#{items[1]}\n" if command == "head"
            if (s.to_i <= items[0].to_i) && (items[0].to_i <= e.to_i) then
              if id.nil? then
                @body << line
              else
                @body << line if items[1] == id
              end
            end
          }
        end
      end

    # update
    elsif path =~ %r!^update/([0-9A-Za-z_]+)/(\d+)/([0-9A-Za-z]+)/([^:]*):(\d+)(.*)! then
      file, stamp, id, host, nport, path = $1, $2, $3, $4, $5, $6
      host = ENV["REMOTE_ADDR"] if host == ""
      node = Node.new(host, nport, path)
      if update(file, stamp, id, node) then
        @body << "OK\n"
      end
    end
  end

  #
  # relay update infomation
  #
  def update(file, stamp, id, node)
    Util.printtrace(file, stamp, id, node) ## debug
    datfile    = "#{$conf["datadir"]}/#{file}.dat"

    if (Time.new.to_i - stamp.to_i) > ($conf["saveupdate"] * 60) then
      return false
    end

    idline = "#{stamp}<>#{id}<>#{file}\n"

    updates = (IO.readlines($conf["updatelist"]) rescue [])
    updates.each {|line|
      return true if line == idline
    }

    Util.addUpdate(stamp, id, file)

    flag = false
    if ! FileTest.file?(datfile) then
      # ファイルが存在しない --> 他ノードにupdate通知
      Cache.tellupdate(file, stamp, id, node)
      return false
    elsif FileTest.size?(datfile) then
      # サイズがある --> 
      flag = Cache.getData(file, stamp, id, node)
    else
      # サイズがゼロ --> 
      Cache.getRegion(file, node)
      flag = FileTest.size?(datfile)
    end

    if flag then
      Cache.tellupdate(file, stamp, id, nil)
      unless (NodeList.include(node) || (NodeList.count() > $conf["nodes"])) then
        NodeList.join(node)
      end
      return true
    else
      removeUpdate(idline)
      return false
    end
  end

  #
  # check client == node ?
  #
  def checkAddress(host, ip)
    Util.printtrace(host, ip) ## debug

    return true if host == ip

    require 'Socket'
    $conf["retry"].times {|i|
      addr = Socket.unpack_sockaddr_in(Socket.gethostbyname(host)[3])[1]
      if addr != "" then
        return addr == ip
      end
    }
    return false
  end

  #
  # remove from update list
  #
  def removeUpdate(idline)
    Util.printtrace(idline) ## debug

    lists = (IO.readlines($conf["updatelist"]) rescue [])
    return if lists.empty?

    save = lists.reject! {|line| line == idline}
    return unless save

    Util.lock()
    File.open($conf["updatelist"], "w") {|f|
      f << lists
    }
    Util.unlock()
  rescue Errno.ENOENT, Errno.EACCES
    raise "#{$conf["updatelist"]}: failed to open. #{$!}"
  end
end
