
require "config"
require "node"
require "util"

module NodeList
  module_function        # public function

  #
  # show all nodes
  #
  def list()
    Util.printtrace() ## debug

    return [] unless FileTest.exist?($conf["nodelist"])

    nodes = []
    IO.foreach($conf["nodelist"]) {|line|
      line.chomp!
      nodes << Node.new(line)  # node = host, port, path
    }
    return nodes
  end


  #
  # show randmized all nodes
  #
  def all()
    Util.printtrace() ## debug

    nodes = list()
    nodes.size.downto(1) {|i| # size --> 1
      n = rand(i)             # 0 <= n < i
      nodes[i-1], nodes[n] = nodes[n], nodes[i-1]  # swap
    }
    return nodes
  end

  #
  # suggest random node
  #
  def random()
    Util.printtrace() ## debug

    nodes = list()
    if nodes.size > 0 then
      n = rand(nodes.size)     # 0 <= n < i
      return nodes[n]
    else
      return nil
    end
  end

  #
  # who am i
  #
  def myself()
    Util.printtrace() ## debug

    port = ENV["SERVER_PORT"].dup
    path = $conf["path"].dup

    if ENV["SERVER_NAME"] != "" then
      return Node.new(ENV["SERVER_NAME"], port, path)
    end

    nodes = all()
    nodes.each {|node|
      buf = node.talk("/ping") # 送信元アドレス buf[1]
      if (buf.size >= 2) && (buf[0] == "PONG") then
        return Node.new(buf[1], port, path)
      end
    }
    return nil
  end

  #
  # add node to list
  #
  def add(node)
    Util.printtrace(node) ## debug

    str = node.toString() + "\n"
    unless include(node) then
      Util.lock()
      begin
        File.open($conf["nodelist"], "a") {|f|
          f << str
        }
        Util.debug("★Node追加 : ", node)
      ensure
        Util.unlock()
      end
    end
  rescue Errno.ENOENT, Errno.EACCES
    raise "#{$conf["nodelist"]}: open failed. #{$!}"
  end

  #
  # include node
  #
  def include(node)
    Util.printtrace(node) ## debug

    str = node.toString()
    nodes = list()
    nodes.each {|node|
      return true if node.toString() == str
    }
    return false
  end

  #
  # count node
  #
  def count()
    Util.printtrace() ## debug

    return list().size
  end

  #
  # remove node from list
  #
  def remove(node) 
    Util.printtrace(node) ## debug
    str = node.toString() + "\n"

    nodelist = (IO.readlines($conf["nodelist"]) rescue [])   # read all lines
    return if nodelist.empty?

    save = nodelist.reject! {|line| line == str }
    return unless save

    Util.debug("★Node削除 : ", node)

    Util.lock()
    begin
      File.open($conf["nodelist"], "w") {|f|
        f << nodelist
      }
    ensure
      Util.unlock()
    end
  end

  #
  # ping all nodes
  #
  def pingall()
    Util.printtrace() ## debug

    nodes = all()
    nodes.each {|node|
      remove(node) unless node.ping() # 返答なしのNodeはリストから削除
    }
  end

  #
  # join network
  #
  def join(node)
    Util.printtrace(node) ## debug

    mys = myself()
    if mys.nil? || (mys.toString == node.toString) then
      Util.debug("  ★MyselfNode : ", myself, node) ## debug
      return
    end

    port = ENV["SERVER_PORT"].dup
    count = 0
    loop {
      buf = node.join()
      if (buf.size == 1) && (buf[0] == "WELCOME") then
        add(node)
        break
      elsif (count+1) > $conf["retrySearch"] then
        break
      elsif (buf.size == 2) && (buf[0] == "WELCOME") then
        add(node)
        node = Node.new(buf[1])  # 紹介されたNode
        Util.debug("  ★紹介されたNode : ", node) ## debug
      else
        break
      end
      count += 1
    }
  end

  #
  # connect initial node
  #
  def init()
    Util.printtrace() ## debug

    port = ENV["SERVER_PORT"].dup
    initNodes = $conf["initNode"]

    if ENV["X_INIT_NODE"] then
      s = ENV["X_INIT_NODE"].split(/\+/)
      initNodes.unshift(s)  # add to head
    end

    initNodes.each {|node_str|
      node = Node.new(node_str)
      if node.ping() then   # ping
        NodeList.join(node) # join
        break
      end
    }

    mys = myself()
    remove(mys) if mys

    nodes = all()   # Array
    done  = {}      # Hash
    return if nodes.size == 0

    nodes.each {|node|
      done[node.toString()] = 1
    }

    loop {
      node = nodes[rand(nodes.size)]
      buf  = node.talk("/node")
      str = ""

      if (buf.size > 0) && (buf[0] =~ /:\d+\/[^: ]+$/) then
        str = buf[0]
        node = Node.new(str)
        nodes << node unless done[str]
      end

      if done[str] then
        done[str] += 1
      else
        done[str] = 1
      end
      if (done[str] > $conf["retrySearch"]) ||
         (nodes.size >= $conf["nodes"]) then
        break
      end
    }

    nodes.each {|node|
      NodeList.join(node) unless include(node)
    }

    remove(mys) if mys

    if count() >= $conf["nodes"] then
      node.bye()
      remove(node)
    end

    while (count() > $conf["nodes"]) do
      node = random()
      node.bye()
      remove(node)
    end
  end
end
